#!/usr/bin/env python
# -*- coding: utf-8 -*-

# dvhaedit_app.py
"""
Script to start DVHA DICOM Editor
"""
# Copyright (c) 2020 Dan Cutright
# This file is part of DVH Analytics, released under a BSD license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-DICOM-Editor

import dvhaedit.main

if __name__ == "__main__":
    dvhaedit.main.start()
